// SpacePartitionProxy.cpp: implementation of the SpacePartitionProxy class.
//
//////////////////////////////////////////////////////////////////////

#include "SpacePartitionProxy.h"
#include "IM1Decoders/AFX/AFX.h"
#include "IM1Decoders/AFX/SpacePartitionDecoder.h"

#include "opengl/AFX/BitWrapperProxy.h"
#include "opengl/group/TransformProxy.h"
#include "opengl/group/SwitchProxy.h"

#include "opengl/base/OpenGLPresenter.h"
#include "Presenter.h"

#pragma warning(disable:4786)

static int nCount=0;

#define MAX_BSP_OBJ_TO_DISPLAY 25

static bool bAddedToGraph = false;

//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

SpacePartitionProxy::SpacePartitionProxy(SpacePartition* spacepartition)
  : AfxStreamProxy(spacepartition)
{
  m_pSpacePartition = new SPart();
  m_pSpacePartition->SetMediaObject((MediaObject*)GetSpacePartitionNode());
  m_bInit = false;
  m_lastTime = 0;
  bWssAdded = false;
}

SpacePartitionProxy::~SpacePartitionProxy(void)
{
	delete m_pSpacePartition;
	m_pSpacePartition=NULL;	
}


/*---------------- virtual functions -----------------------------*/
BOOL SpacePartitionProxy::Select(OpenGLVisualRenderer* pRenderer, DWORD dwTime, unsigned int& depth)
{
  GLuint selectBuffer[4];

  ::glSelectBuffer(4, selectBuffer);
  ::glRenderMode(GL_SELECT);

//  Render(pRenderer, dwTime);

  if (::glRenderMode(GL_RENDER) > 0)
    if (depth >= selectBuffer[1]) {
      depth = selectBuffer[1];

      return TRUE;
    }

  return FALSE;
}

BOOL SpacePartitionProxy::RenderFromStream(MediaStream* poStream, int id, DWORD dwTime)
{
	// it is called by BitWrapper
	// it should retrieve from the decoder the data and initialise the fields

	SpacePartition *pObject = GetSpacePartitionNode();
    BitWrapper *pParent = (BitWrapper *)pObject->GetParent();
    ASSERT(pParent != NULL);

	// first time get the decoder specific info from the stream
	if (!m_bInit)
	{
		m_bInit = true;
	}

	// get the stream
	MediaStream * pStream = (MediaStream *)pParent->GetStream();
	ASSERT(pStream != NULL);

	// get the decoded data buffer
	int rnSize, rnStatus;
	DWORD rdvTime;
	BOOL rbEOU;
	BYTE * pOutput = (BYTE *)pStream->Fetch(rnSize, rdvTime, rbEOU, rnStatus);


	//Copy the decoded mesh into the composition buffer
	//DataExchange();
	
	// update time
	if (rdvTime > m_lastTime) {
		m_lastTime = rdvTime;
		return TRUE;
	}

	return FALSE;
}


BOOL SpacePartitionProxy::RenderFromBuffer(SFString *buffer, int id)
{
	SpacePartDecoderLib SpacePartDec;
	SpacePartDec.SetSPart(m_pSpacePartition);
	SpacePartDec.DecodeFromBuffer (buffer->GetValue(),buffer->GetLength());
	return TRUE;
}


void SpacePartitionProxy::SetupDecoderData(BitWrapperProxy *pParent)
{
	SpacePartition *pObject = GetSpacePartitionNode();

	BitWrapper* poBitWrapper = pParent->GetBitWrapper();
	AFXDecoder *pDecoder = (AFXDecoder *)poBitWrapper->GetDecoder();
	ASSERT(pDecoder != NULL);
	
	SpacePartitionDecoder *spDec = (SpacePartitionDecoder *)pDecoder->GetDecoder();
	ASSERT(spDec != NULL);

	spDec->SetRenderingObject(this);
}


void SpacePartitionProxy::RenderDescription(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
//  FootPrintSet* pFootPrintSet = (FootPrintSet*) GetMediaObject();
  GLfloat color[] = { 0.8f, 0.8f, 0.8f, 0.75f};

  glColor4fv(color);
  ::glPushMatrix();
  
  ::glScalef(1.0f, 1.0f, 1.0f);
  ::glTranslatef(0.0f, -2.0f, 0.0f);

  RenderDescriptionBox(pRenderer, "SpacePartition");

  ::glTranslatef(0.0f, -1.0f, 0.0f);

/*  ZString text("FootPrint (");
  char nbchildren[10];
  _itoa( pFootPrintSet->GetChildrenCount(),nbchildren,10);
  text+=nbchildren;
  text+=")";


  RenderDescriptionBox(pRenderer,(LPCSTR)"SP");
*/  
  ::glPopMatrix();
  ::glPopMatrix();
}



/*
BOOL SpacePartitionProxy::Render(const Matrix& rTransformation, const CullFlag& rCullFlag,
		   OpenGLVisualRenderer* pRenderer, DWORD dwTime)
*/
BOOL SpacePartitionProxy::Render(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
    // called when the node is encoded with BIFS

	// Get the viewpoint position
	Viewpoint* pActiveViewpoint = GetMediaObject()->GetRoot()->GetBoundViewpoint();
	SF<Vec3f>& position = *(SF<Vec3f>*) pActiveViewpoint->GetField(4);

	SpacePartition* pSP = (SpacePartition*) GetMediaObject();
	SpacePartition *pNode = GetSpacePartitionNode();

	Root* pSceneRoot = NULL;  
	OpenGLPresenter* pPresenter = NULL;
	int nModels = 0;
	float* translations = NULL;
	
	pPresenter = pRenderer->GetPresenter();
	if (pPresenter != NULL)
	{
		pSceneRoot = pPresenter->GetScene();
	}

	switch (m_pSpacePartition->m_SPtype)
	{
	case 0: // BSP
		RenderBSP(pRenderer, dwTime);
		break;

	case 1: // Cell/Portal
		RenderCellPortal(pRenderer, dwTime);
		break;

	case 2: // PVS
		RenderPVS(pRenderer, dwTime);
		break;

	case 3: // footprint
		RenderHDesc(pRenderer, dwTime);
		break;

	case 4: // wss
		RenderWSS(pRenderer, dwTime);
		break;
	}



  return TRUE;
}


BOOL SpacePartitionProxy::RenderBSP(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
	Root* pSceneRoot = NULL;  
	OpenGLPresenter* pPresenter = NULL;

	pPresenter = pRenderer->GetPresenter();
	if (pPresenter != NULL)
	{
		pSceneRoot = pPresenter->GetScene();
	}

	// Get the viewpoint position
	Viewpoint* pActiveViewpoint = GetMediaObject()->GetRoot()->GetBoundViewpoint();
	SF<Vec3f>& position = *(SF<Vec3f>*) pActiveViewpoint->GetField(4);
	
	std::list<Switch*>::iterator itobj;
	for (itobj=m_DisplayList.begin(); itobj!=m_DisplayList.end(); itobj++)
	{
		Switch* pSwitch = *itobj;
		if (pSwitch != NULL) pSwitch->whichChoice = -1;
	}
	m_DisplayList.empty();
	m_DisplayList.clear();

	// current cell
	
	ZArray<sint32> ObjIDs;
	m_pSpacePartition->m_BSPTree->SelectObjects(position.GetValue().x, position.GetValue().y, position.GetValue().z, &ObjIDs);

	int nobjcount = ObjIDs.GetLength();
	for (int iobj=0; iobj<10; iobj++)
	{
		sint32 id = ObjIDs.ElementAt(iobj);

		char nodename[50];
		sprintf(nodename, "s%d", id);
		Switch* pSwitch=(Switch*)pSceneRoot->FindNode(LPCSTR(nodename));
		if (pSwitch != NULL)
		{
			pSwitch->whichChoice = 0;
			m_DisplayList.push_back(pSwitch);
		}
		
/*		MediaObject* pObj = pNode->GetChild(id);
		MediaObjectProxy * pProxy = (MediaObjectProxy *) (pObj->GetRenderingObject());

		pSceneRoot->AddChild(pObj);
*/
	}

	return true;
}


BOOL SpacePartitionProxy::RenderCellPortal(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
	Root* pSceneRoot = NULL;  
	OpenGLPresenter* pPresenter = NULL;

	// Get the viewpoint position
	Viewpoint* pActiveViewpoint = GetMediaObject()->GetRoot()->GetBoundViewpoint();
	SF<Vec3f>& position = *(SF<Vec3f>*) pActiveViewpoint->GetField(4);

	// detect in which cell observer is
	CPPartition* pCPPartition = m_pSpacePartition->m_CPPartition;
	if (pCPPartition == NULL)
		return true;

	int nModels = 0;
	float* translations = NULL;
	
	pPresenter = pRenderer->GetPresenter();
	if (pPresenter != NULL)
	{
		pSceneRoot = pPresenter->GetScene();
	}

	CPCell* pCurrentCell = pCPPartition->GetCurrentCell(position.GetValue().x, position.GetValue().y, position.GetValue().z);
	if (pCurrentCell == NULL) return true;

	// has cell changed??
	if (pCPPartition->m_pCurrentCell == pCurrentCell)
	{
		// unchanged do nothing
		return true;
	}

	// save new current cell
	pCPPartition->m_pCurrentCell = pCurrentCell;
	
	// remove old objects from graph
	std::list<Switch*>::iterator itobj;
	for (itobj=m_DisplayList.begin(); itobj!=m_DisplayList.end(); itobj++)
	{
		Switch* pSwitch = *itobj;
		if (pSwitch != NULL) pSwitch->whichChoice = -1;
	}
	m_DisplayList.empty();
	m_DisplayList.clear();

	// current cell
	char nodename[50];
	sprintf(nodename, "Cell_%d", pCurrentCell->m_nID);
	Switch* pSwitch=(Switch*)pSceneRoot->FindNode(LPCSTR(nodename));
	if (pSwitch != NULL)
	{
		pSwitch->whichChoice = 0;
		m_DisplayList.push_back(pSwitch);
	}

	// visible cells
	std::list<int> listCells = pCPPartition->GetVisibleCPCells(pCurrentCell);
	std::list<int>::iterator itid;
	for (itid=listCells.begin(); itid!=listCells.end(); itid++)
	{
		int nid = *itid;
		sprintf(nodename, "Cell_%d", nid);
		Switch* pSwitch=(Switch*)pSceneRoot->FindNode(LPCSTR(nodename));
		if (pSwitch != NULL)
		{
			pSwitch->whichChoice = 0;
			m_DisplayList.push_back(pSwitch);
		}
	}

	return true;
}


BOOL SpacePartitionProxy::RenderPVS(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
	// Get the viewpoint position
	Viewpoint* pActiveViewpoint = GetMediaObject()->GetRoot()->GetBoundViewpoint();
	SF<Vec3f>& position = *(SF<Vec3f>*) pActiveViewpoint->GetField(4);

	// detect in which cell observer is
	PvsPartition* pPvsPartition = m_pSpacePartition->m_PVSPart;

	SpacePartition *pNode = GetSpacePartitionNode();

	Root* pSceneRoot = NULL;  
	OpenGLPresenter* pPresenter = NULL;
	int nModels = 0;
	float* translations = NULL;
	
	pPresenter = pRenderer->GetPresenter();
	if (pPresenter != NULL)
	{
		pSceneRoot = pPresenter->GetScene();
	}

	if (pPvsPartition->m_IsRegular == false)
	{
//		ReadIFS();
	}

	PvsCell* pCell = pPvsPartition->GetPvsCell(position.GetValue().x, position.GetValue().y, position.GetValue().z);
	if (pCell == NULL) return true;

	// has cell changed??
	if (pPvsPartition->m_CurrentCell == pCell)
	{
		// unchanged do nothing
		return true;
	}

	// save new current cell
	pPvsPartition->m_CurrentCell = pCell;
	
	// remove old objects from graph
	std::list<Switch*>::iterator itobj;
	for (itobj=m_DisplayList.begin(); itobj!=m_DisplayList.end(); itobj++)
	{
		Switch* pSwitch = *itobj;
		if (pSwitch != NULL) pSwitch->whichChoice = -1;
	}
	m_DisplayList.empty();
	m_DisplayList.clear();

	// add visible objects
	std::list<int>* pObjID = pPvsPartition->GetObjIDs(pCell);
	std::list<int>::iterator itid;
	for (itid=pObjID->begin(); itid!=pObjID->end(); itid++)
	{
		int id = *itid;
		char nodename[50];
		sprintf(nodename, "G_%06d", id+1);
		Switch* pSwitch=(Switch*)pSceneRoot->FindNode(LPCSTR(nodename));
//		MediaObject* pObj = pNode->GetChild(id);
//		pSceneRoot->AddChild(pObj);
		if (pSwitch != NULL) 
		{
			pSwitch->whichChoice = 0;
			m_DisplayList.push_back(pSwitch);
		}
	}

	return true;
}


BOOL SpacePartitionProxy::RenderHDesc(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
//	RenderHDescSphere(NULL);
//	return true;

	HDescTree* pTree = m_pSpacePartition->m_HDescTree;
	if (pTree == NULL) return false;

	nCount=0;

/*	int nnodes = pTree->m_SortedNodes.size();
	SORTED_NODES::iterator itnodes;
	for (itnodes=pTree->m_SortedNodes.begin(); itnodes!=pTree->m_SortedNodes.end(); itnodes++)
	{
		std::pair<class HDescNode * const,int> pPair = *itnodes;
		HDescNode* pNode = pPair.first;
		int i = pPair.second;
	}
	*/

	HDescNode * pNode = pTree->m_pRoot;
	return RenderHDescNode(pNode);
}


BOOL SpacePartitionProxy::RenderHDescNode(HDescNode* pNode)
{
	if (pNode == NULL) return true;

	// render sphere
	RenderHDescSphere(pNode);
/*
	// render brother node
	RenderHDescNode(pNode->GetBrother());
*/
	// render children 
	std::vector<HDescNode*>::iterator it;
	for (it=pNode->m_Children.begin(); it!=pNode->m_Children.end(); it++)
	{
		HDescNode* pChild = *it;
		RenderHDescNode(pChild);
	}

	return true;
}

BOOL SpacePartitionProxy::RenderHDescSphere(HDescNode* pNode)
{
	nCount++;

	::glEnable(GL_LIGHTING);
	::glEnable(GL_COLOR_MATERIAL);

	double radius = 15; //pNode->m_radius;
	float xt = pNode->m_fGcX;
	float yt = pNode->m_fGcY;
	float zt = pNode->m_fGcZ;
	int lats = 10;
	int longs = 10;

	int i, j;
	for(i = 0; i <= lats; i++) {
		double lat0 = M_PI * (-0.5 + (double) (i - 1) / lats);
		double z0  = sin(lat0);
		double zr0 =  cos(lat0);

		double lat1 = M_PI * (-0.5 + (double) i / lats);
		double z1 = sin(lat1);
		double zr1 = cos(lat1);
    
		bool blight = glIsEnabled(GL_LIGHTING);
		float mat0[] = {0.5f,0.5f,0.5f,0.5f};
		float mat1[] = {0.f,1.f,0.f};
		float mat2[] = {0.5f,1.f,0.5f};
		glMaterialfv(GL_FRONT_AND_BACK, GL_AMBIENT, mat0);
		glMaterialf(GL_FRONT_AND_BACK, GL_SHININESS, 2.0f);
		glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, mat1);
		glMaterialfv(GL_FRONT_AND_BACK, GL_SPECULAR, mat2);

        glBegin(GL_QUAD_STRIP);
        for(j = 0; j <= longs; j++) {
			double lng = 2 * M_PI * (double) (j - 1) / longs;
            double x = cos(lng);
            double y = sin(lng);
    
            glNormal3f(x * zr0, y * zr0, z0);
            glVertex3f(xt + radius * x * zr0, yt + radius * y * zr0, zt + radius * z0);
            glNormal3f(x * zr1, y * zr1, z1);
            glVertex3f(xt + radius * x * zr1, yt + radius * y * zr1, zt + radius * z1);
        }
        glEnd();
	}

	return true;
}


BOOL SpacePartitionProxy::RenderWSS(OpenGLVisualRenderer* pRenderer, DWORD dwTime)
{
	// pass WssPartition + render child wss node
/*	if (bWssAdded == true)
		return true;
*/
	Root* pSceneRoot = NULL;  
	OpenGLPresenter* pPresenter = NULL;
	pPresenter = pRenderer->GetPresenter();
	if (pPresenter != NULL)
	{
		pSceneRoot = pPresenter->GetScene();
	}

	WssPartition* pWssPartition = m_pSpacePartition->m_WSSPart;

	SpacePartition *pNode = GetSpacePartitionNode();
	int nbchildren=pNode->GetChildrenCount();
	if ( (bWssAdded == false) && (nbchildren!=0) )
	{		
		for (int j=0; j<nbchildren; j++)
		{
			MediaObject* pObject=pNode->GetChild(j);
			pSceneRoot->AddChild(pObject);
//			((WaveletSubdivisionSurfaceProxy *) ((pObject->GetRenderingObject()))->Render(pRenderer,dwTime);
		}
	}
	bWssAdded = true;

	std::list<WssVolume*>::iterator itvolume;
	for (itvolume=pWssPartition->m_Volumes.begin(); itvolume!=pWssPartition->m_Volumes.end(); itvolume++)
	{
		//-- render volume --
		WssVolume* pVol = *itvolume;


		// compute volume coordinates
		float * pVolumeCoord;
		pVolumeCoord = (float*)calloc(3*8, sizeof(float));
		// pt 0
		pVolumeCoord[0] = pVol->m_centerX-pVol->m_dX/2;
		pVolumeCoord[1] = pVol->m_centerY-pVol->m_dY/2;
		pVolumeCoord[2] = pVol->m_centerZ-pVol->m_dZ/2;
		// pt 1
		pVolumeCoord[3] = pVol->m_centerX+pVol->m_dX/2;
		pVolumeCoord[4] = pVol->m_centerY-pVol->m_dY/2;
		pVolumeCoord[5] = pVol->m_centerZ-pVol->m_dZ/2;
		// pt 2
		pVolumeCoord[6] = pVol->m_centerX+pVol->m_dX/2;
		pVolumeCoord[7] = pVol->m_centerY+pVol->m_dY/2;
		pVolumeCoord[8] = pVol->m_centerZ-pVol->m_dZ/2;
		// pt 3
		pVolumeCoord[9] = pVol->m_centerX-pVol->m_dX/2;
		pVolumeCoord[10] = pVol->m_centerY+pVol->m_dY/2;
		pVolumeCoord[11] = pVol->m_centerZ-pVol->m_dZ/2;

		// pt 4
		pVolumeCoord[12] = pVol->m_centerX-pVol->m_dX/2;
		pVolumeCoord[13] = pVol->m_centerY-pVol->m_dY/2;
		pVolumeCoord[14] = pVol->m_centerZ+pVol->m_dZ/2;
		// pt 5
		pVolumeCoord[15] = pVol->m_centerX+pVol->m_dX/2;
		pVolumeCoord[16] = pVol->m_centerY-pVol->m_dY/2;
		pVolumeCoord[17] = pVol->m_centerZ+pVol->m_dZ/2;
		// pt 6
		pVolumeCoord[18] = pVol->m_centerX+pVol->m_dX/2;
		pVolumeCoord[19] = pVol->m_centerY+pVol->m_dY/2;
		pVolumeCoord[20] = pVol->m_centerZ+pVol->m_dZ/2;
		// pt 7
		pVolumeCoord[21] = pVol->m_centerX-pVol->m_dX/2;
		pVolumeCoord[22] = pVol->m_centerY+pVol->m_dY/2;
		pVolumeCoord[23] = pVol->m_centerZ+pVol->m_dZ/2;


//		unsigned int* pVolumeCoordInd;
//		pVolumeCoordInd = (unsigned int*)calloc(4*6, sizeof(unsigned int));
		unsigned int pVolumeCoordInd[] = { 0,1,2,3,
								1,5,6,2,
								3,7,4,0,
								4,7,6,5,
								2,6,7,3,
								0,4,5,1 };
							

		glCullFace(GL_FRONT_AND_BACK);

	
		// Enable vertex and normal arrays
		::glEnableClientState(GL_VERTEX_ARRAY);

		// Set pointers to buffers
		::glVertexPointer(3, GL_FLOAT, 0, pVolumeCoord);

		// Render
		::glDrawElements(GL_QUADS, 24/*_pWaveSurf->GetCoordIndexLength() * 3 / 4*/, GL_UNSIGNED_INT, pVolumeCoordInd);
		::glDisableClientState(GL_VERTEX_ARRAY);
	}
	return true;
}

