// BSP.h: interface for the BSP class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_BSP_H__1967247C_039D_4314_91E5_D232534ED9A8__INCLUDED_)
#define AFX_BSP_H__1967247C_039D_4314_91E5_D232534ED9A8__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

//#include "IDHashTable.h"
#include "ztl.h"

//! the 1 byte integer
typedef signed char sint8;
typedef unsigned char uint8;

//! the 2 bytes integer
typedef signed short sint16;
typedef unsigned short uint16;

//! the 4 bytes integer
typedef signed long sint32;
typedef unsigned long uint32;


#include <list>


class BSPNode
{
public:
	BSPNode(void) : m_nIndex(-1), m_nParentIndex(-1), m_a(0), m_b(0), m_c(0), m_d(0), m_nFrontIndex(-1), m_nOverlapIndex(-1), m_nBackIndex(-1), m_nFrontObjCount(0), m_lFrontObjID(NULL), m_nOverlapObjCount(0), m_lOverlapObjID(NULL), m_nBackObjCount(0), m_lBackObjID(NULL), m_pParentNode(NULL), m_pFrontNode(NULL), m_pOverlapNode(NULL), m_pBackNode(NULL) {};
	BSPNode(sint32 nIndex, sint32 nParentIndex, float a, float b, float c, float d, sint32 nFrontIndex, sint32 nOverlapIndex, sint32 nBackIndex) : m_nIndex(nIndex), m_nParentIndex(nParentIndex), m_a(a), m_b(b), m_c(c), m_d(d), m_nFrontIndex(nFrontIndex), m_nOverlapIndex(nOverlapIndex), m_nBackIndex(nBackIndex), m_nFrontObjCount(0), m_lFrontObjID(NULL), m_nOverlapObjCount(0), m_lOverlapObjID(NULL), m_nBackObjCount(0), m_lBackObjID(NULL), m_pParentNode(NULL), m_pFrontNode(NULL), m_pOverlapNode(NULL), m_pBackNode(NULL) {};

	void SetParentNode(BSPNode* pNode) { m_pParentNode = pNode; };
	void SetFrontNode(BSPNode* pNode) { m_pFrontNode = pNode; };
	void SetOverlapNode(BSPNode* pNode) { m_pOverlapNode = pNode; };
	void SetBackNode(BSPNode* pNode) { m_pBackNode = pNode; };


	bool isRoot(void) { return (m_nParentIndex == -1); };
public:
	~BSPNode(void);

	// AU
	sint32	m_nIndex;
	sint32	m_nParentIndex;
	float	m_a;
	float	m_b;
	float	m_c;
	float	m_d;
	sint32	m_nFrontIndex;
	sint32  m_nOverlapIndex;
	sint32	m_nBackIndex;
	uint32  m_nFrontObjCount;
	sint32* m_lFrontObjID;
	uint32  m_nOverlapObjCount;
	sint32* m_lOverlapObjID;
	uint32  m_nBackObjCount;
	sint32* m_lBackObjID;

	// parent and children nodes
	BSPNode*	m_pParentNode;
	BSPNode*	m_pFrontNode;
	BSPNode*	m_pOverlapNode;
	BSPNode*	m_pBackNode;
};


class BSPTree
{
public:
	BSPTree(void);
public:
	~BSPTree(void);

public:
	void AddNode(BSPNode* pNode);
	int SelectViewpoint(int maxdepth, float x, float y, float z, ZArray<sint32>* pObjID);
	int SelectObjects(float x, float y, float z, ZArray<sint32>* pObjID);
	int SelectFrustrum(int maxdepth, float* X, float* Y, float* Z, ZArray<sint32>* pObjID);
	int SelectOrientation(float x, float y, float z, float vX, float vY, float vZ, ZArray<sint32>* pObjID);
	int Parse(int depth, int maxdepth, BSPNode* pNode, float x, float y, float z, ZArray<sint32>* pObjID);
	int Parse(int depth, int maxdepth, BSPNode* pNode, float* x, float* y, float* z, ZArray<sint32>* pObjID);
	int ParseFrustrum(BSPNode* pNode, float* x, float* y, float* z, ZArray<sint32>* pObjID);
	int FillAllChildren(BSPNode* pNode, ZArray<sint32>* pObjID);

	// basic ordering algo
	int ParseObjects(int depth, BSPNode* pNode, float x, float y, float z, ZArray<sint32>* pObjID);
	int ParseFrontObjects(int depth, BSPNode* pNode, float x, float y, float z, ZArray<sint32>* pObjID);
	int ParseOverlapObjects(int depth, BSPNode* pNode, float x, float y, float z, ZArray<sint32>* pObjID);
	int ParseBackObjects(int depth, BSPNode* pNode, float x, float y, float z, ZArray<sint32>* pObjID);

private:
	BSPNode*		m_pRoot;
	BSPNode**		m_pNodesByID;
	BSPNode**		m_pOrphansByID;

public:
	int				m_nTotalNodes;

	std::list<int>	toto;
};

#endif // !defined(AFX_BSP_H__1967247C_039D_4314_91E5_D232534ED9A8__INCLUDED_)
