// CPPartition.cpp: implementation of the CPPartition class.
//
//////////////////////////////////////////////////////////////////////

#include "CPPartition.h"
#include <math.h>



bool CPCell::IsInsideCell(float x, float y, float z)
{
	if (x > m_xcenter + m_dx/2) return false;
	if (x < m_xcenter - m_dx/2) return false;
	if (y > m_ycenter + m_dy/2) return false;
	if (y < m_ycenter - m_dy/2) return false;
	if (z > m_zcenter + m_dz/2) return false;
	if (z < m_zcenter - m_dz/2) return false;

	return true;
}


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

CPPartition::CPPartition(void)
{
	m_Cells.clear();
	m_pCurrentCell = NULL;
}

CPPartition::~CPPartition(void)
{
}

std::list<int> CPPartition::GetVisibleCPCells(CPCell* pCell)
{
	if (pCell->m_State == CP_BUILT)
	{
		return pCell->m_listCellIDs;
	}
	else
	{
		int nBytes = (int)ceil((float)((float)m_nCellCount/8));
		int iObj=0;
		for (int iByte=0; iByte<nBytes; iByte++) 
		{
			unsigned char Byte = (unsigned char) pCell->m_bitarray[iByte];
			for (int iBit=0; iBit<8; iBit++) 
			{
				unsigned char flag = (unsigned char)pow((double)2, (int)iBit);
				if (flag & Byte) {
					pCell->m_listCellIDs.push_back(iObj);
				}
				else {
					// skip
				}
				iObj++;
			}
		}
		delete(pCell->m_bitarray); pCell->m_bitarray=NULL;
		pCell->m_State = CP_BUILT;

		return pCell->m_listCellIDs;
	}
}

CPCell* CPPartition::GetCurrentCell(float x, float y, float z)
{
	CPCell* pCellOUT = NULL;

	if (m_pCurrentCell != NULL)
	{
		// modified?
		if (m_pCurrentCell->IsInsideCell(x, y, z))
			return m_pCurrentCell;
	}

	{
		std::list<CPCell*>::iterator itcell;
		for (itcell=m_Cells.begin(); itcell!=m_Cells.end(); itcell++)
		{
			CPCell* pCell = *itcell;
			if (pCell != NULL) {
				if (pCell->IsInsideCell(x, z, y)) {
//					m_pCurrentCell = pCell;
//					pCellOUT = m_pCurrentCell;
					pCellOUT = pCell;
				}
			}
		}
		// parse cells
/*		CPCell* pCell = new CPCell;
		pCell->m_xcenter = x-5.0;
		pCell->m_ycenter = z-5.0;
		pCell->m_zcenter = y;

		std::set<CPCell*, lt_Cell>::iterator it1 = m_SetCells.lower_bound(pCell);
		std::set<CPCell*, lt_Cell>::iterator it2 = m_SetCells.upper_bound(pCell);
//		it2--;
		if (it1 != m_SetCells.end())
		{
			//delete(pCell);
			m_pCurrentCell = *it1;
			if (m_pCurrentCell->IsInsideCell(x, y, z))
				pCellOUT = m_pCurrentCell;
		}
	}
*/
	}
	return pCellOUT;
}


void CPPartition::AddCell(CPCell* pCell)
{
	m_Cells.push_back(pCell);
}