// HDescTree.h: interface for the HDescTree class.
//
//////////////////////////////////////////////////////////////////////

#if !defined(AFX_HDESCTREE_H__10B69F30_6146_4CCC_9EEE_38B50BC4C259__INCLUDED_)
#define AFX_HDESCTREE_H__10B69F30_6146_4CCC_9EEE_38B50BC4C259__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#pragma warning(disable: 4786)

#include <list>
#include <utility>
#include <map>
#include <vector>



class HDescNode
{
public:
	HDescNode();
	HDescNode(int index) { m_index = index; };
	HDescNode(float fGcX, float fGcY, float fGcZ, float fMetricError, int s4Radius= 0);

	virtual ~HDescNode();

	inline int GetIndex(void) { return m_index; };
	inline void SetIndex(int index) { m_index = index; };
	inline int GetLevel(void) { return m_index; };
	inline void SetLevel(int level) { m_level = level; };
	inline int GetNbMissingChildren(void) { return m_nbMissingChildren; };
	inline void SetNbMissingChildren(int nb) { m_nbMissingChildren = nb; };
	inline int GetRadius(void) { return m_radius; };
	inline void SetRadius(int radius) { m_radius = radius; };
	inline void GetGCCoordinates(float &fGcX, float &fGcY, float &fGcZ) { fGcX=m_fGcX; fGcY=m_fGcY; fGcZ=m_fGcZ; };
	inline void SetGCCoordinates(float fGcX, float fGcY, float fGcZ) { m_fGcX=fGcX; m_fGcY=fGcY; m_fGcZ=fGcZ; };
	inline bool IsRoot() { return (m_index == -1); };

	inline HDescNode* GetParent(void) { return m_pParent; };
	inline void SetParent(HDescNode* pNode) { m_pParent = pNode; };

	inline HDescNode* GetBrother(void) { return m_pBrother; };
	inline void SetBrother(HDescNode* pNode) { m_pBrother = pNode; };

	void AddChild(HDescNode * pNode) { m_Children.push_back(pNode); };

	float	m_fGcX;
	float	m_fGcY;
	float	m_fGcZ;
	float	m_fMetricError;
	int		m_index;
	int		m_level;
	int		m_radius;
	int		m_nbMissingChildren;

	HDescNode*		m_pParent;
	HDescNode*		m_pBrother;
	std::vector<HDescNode*>	m_Children;

};

struct lt_hdescnode
{
	bool operator()(HDescNode* a, HDescNode* b) const
	{
		return (a->m_index < b->m_index);
	}
};

typedef std::map<HDescNode*, int, struct lt_hdescnode> SORTED_NODES;


class HDescTree  
{
public:
	HDescTree();
	virtual ~HDescTree();

	void AddRoot();
	void AddNode(HDescNode* pNode);

	HDescNode* GetNode(int index);

	HDescNode*		m_pRoot;
	SORTED_NODES	m_SortedNodes;

};

#endif // !defined(AFX_HDESCTREE_H__10B69F30_6146_4CCC_9EEE_38B50BC4C259__INCLUDED_)
