
#include "windows.h"
#include "PvsPartition.h"

PvsPartition::PvsPartition(void)
{
	bComposed = false;
	m_CurrentCell = NULL;
}

PvsPartition::~PvsPartition(void)
{
}

std::list<int>* PvsPartition::GetObjIDs(PvsCell* pCell)
{
	if (pCell->GetState() == CELL_STATE_DECODED) {
		return pCell->GetIDs();
	}
	else {
		FillObjIDs(pCell);
		if (pCell->GetState() == CELL_STATE_DECODED) return pCell->GetIDs();
	}

	return NULL;
}

int PvsPartition::FillObjIDs(PvsCell* pCell)
{
	if (pCell == NULL) return 0;
	if (pCell->GetState() == CELL_STATE_DECODED) return pCell->m_pObjID->size();

	std::list<int>* pIDs = new std::list<int>();
	pIDs->empty();

	int nBytes = (m_nObjCount+7)/8;
	int iObj=0;
	for (int iByte=0; iByte<nBytes; iByte++) {
		unsigned char Byte = (unsigned char) pCell->m_pvs[iByte];
		for (int iBit=0; iBit<8; iBit++) {
			unsigned char flag = (unsigned char)pow((double)2, (int)iBit);
			if (flag & Byte) {
				pIDs->push_back(iObj);
			}
			else {
				// skip
			}
			iObj++;
		}
	}

	pCell->m_pObjID = pIDs;
	delete(pCell->m_pvs); pCell->m_pvs=NULL;
	pCell->SetState(CELL_STATE_DECODED);

	return pIDs->size();
}

void PvsPartition::AddPvsCell(PvsCell* pCell)
{
}

PvsCell* PvsPartition::GetPvsCell(float xVP, float yVP, float zVP)
{
	return NULL;
}


PvsPartitionRegular::PvsPartitionRegular(void)
{
//	m_HashCells = new IDHashTable();
	
	m_nCells=0;
}

PvsPartitionRegular::~PvsPartitionRegular(void)
{
}

void PvsPartitionRegular::AddPvsCell(PvsCell* pCell)
{
//	m_HashCells->appendElement(pCell, pCell->m_nID);
	m_CellsByID[pCell->m_nID] = pCell;
	m_nCells++;
}

PvsCell* PvsPartitionRegular::GetPvsCell(float xVP, float yVP, float zVP)
{
	PvsCell* pCell = NULL;

	// determine ID
	if (xVP<m_xmin) return 0;
	if (xVP>m_xmax) return 0;
	if (zVP<m_ymin) return 0;
	if (zVP>m_ymax) return 0;

	int i = (int) (xVP-m_xmin)/m_dx;
	int j = (int) (zVP-m_ymin)/m_dy;
	int nX = (int) ceil((m_xmax-m_xmin)/m_dx);
	int nID = j*nX + i;

//	int nsize=m_HashCells->getHashSize();
//	pCell = (PvsCell*)m_HashCells->getElement(nID);
	pCell = m_CellsByID[nID];

//	m_CurrentCell = pCell;

	return pCell;
}


PvsPartitionIrregular::PvsPartitionIrregular(void)
{

}

PvsPartitionIrregular::~PvsPartitionIrregular(void)
{
}

void PvsPartitionIrregular::AddPvsCell(PvsCell* pCell)
{
//	m_ListCells.push_back(pCell);
	m_ArrayOfCells.push_back(pCell);
}


void PvsPartitionIrregular::BuildCells(void)
{
	int nIdx = m_pIfsIndices->size();
	int iPoly=0;
	int ipos=0;
//	ArrayOf<Point3DF, 10> Poly;
	
	for (int iIdx=0; iIdx<nIdx; iIdx++)
	{
		int Index = m_pIfsIndices->at(iIdx);
		if (Index == -1)
		{
			ipos=0;
			iPoly++;
		}
		else
		{
			if (ipos <= 2)
			{
				Vec3f point = m_pIfsCoords->at(Index);

				PvsCell* pCell = m_ArrayOfCells.at(iPoly);
				((PvsCellIrregular*)pCell)->m_x[ipos] = point.x;
				((PvsCellIrregular*)pCell)->m_y[ipos] = point.y;
	//			((PvsCellIrregular*)pCell)->m_z[ipos] = point.z;
			}

			ipos++;
		}
	}
}


PvsCell* PvsPartitionIrregular::GetPvsCell(float xVP, float yVP, float zVP)
{
	PvsCell* pOutCell = NULL;

	//std::list<PvsCell*>::iterator it;
	int nCells = m_ArrayOfCells.size();
	for (int icell=0; icell<nCells; icell++)
	{
		PvsCellIrregular* pCell = (PvsCellIrregular*)m_ArrayOfCells.at(icell);
		// test if inside this cell
		if (pCell->IsInside(xVP, zVP))
		{
			pOutCell = pCell;
			break;
		}
	}

//	m_CurrentCell = pOutCell;

	return pOutCell;
}

