// SpacePartDecoderLib.cpp: implementation of the SpacePartDecoderLib class.
//
//////////////////////////////////////////////////////////////////////


#include "SpacePartDecoderLib.h"
#include "./BSP/BSPtree.h"
//#include "./CellPortal/CPPartition.h"
#include "./WSS/WssPartition.h"
#include "./PVS/PvsPartition.h"
#include "./HDesc/HDescTree.h"


//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////

SpacePartDecoderLib::SpacePartDecoderLib()
{
	m_pDSI=new SpacePartDecoderSpecificInfo();
}

SpacePartDecoderLib::~SpacePartDecoderLib()
{
	delete ((SpacePartDecoderSpecificInfo*)m_pDSI);
}

BOOL SpacePartDecoderLib :: Setup (AFXExtDescriptor *pSpecificInfo)
{
	m_pDSI->SPtype=((SpacePartDecoderSpecificInfo*)pSpecificInfo)->SPtype;

	// case BSP
	m_pDSI->BSP_NbBitsIndex = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->BSP_NbBitsIndex;
	m_pDSI->BSP_NbBitsCoef = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->BSP_NbBitsCoef;
	m_pDSI->BSP_NbBitsObjCount = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->BSP_NbBitsObjCount;

	// CellPortal
	m_pDSI->CP_NbBitsCellIndex = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->CP_NbBitsCellIndex; 
	m_pDSI->CP_NbBitsPortalIndex = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->CP_NbBitsPortalIndex;
	m_pDSI->CP_NbBitsCellCount = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->CP_NbBitsCellCount;
	m_pDSI->CP_NbBitsPortalCount = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->CP_NbBitsPortalCount;
	m_pDSI->CP_NbBitsCellPortalCount = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->CP_NbBitsCellPortalCount;

	// PVS
	m_pDSI->PVS_NbBitsIndex = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->PVS_NbBitsIndex;
	m_pDSI->PVS_NbBitsObjCount = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->PVS_NbBitsObjCount;
//	m_pDSI->PVS_NbBitsGridParam = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->PVS_NbBitsGridParam;

	// Footprints
	m_pDSI->HDESC_type = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_type;
	m_pDSI->HDESC_rootChildrenRadiusNbBits = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_rootChildrenRadiusNbBits;
	m_pDSI->HDESC_nbChildrenNbBits = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_nbChildrenNbBits;
	m_pDSI->HDESC_nbSubTreesNbBits = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_nbSubTreesNbBits;
	m_pDSI->HDESC_acquisitionPrecision = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_acquisitionPrecision;
	m_pDSI->HDESC_minMetricError = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_minMetricError;
	m_pDSI->HDESC_maxMetricErrorEncodingFunction = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_maxMetricErrorEncodingFunction;
	m_pDSI->HDESC_nbRootChildren = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_nbRootChildren;
	m_pDSI->HDESC_indexNbBits = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_indexNbBits;
	m_pDSI->HDESC_nbNodesInSubTreeNbBits = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_nbNodesInSubTreeNbBits;
	m_pDSI->HDESC_nbNodesOnFirstLevelOfSubTreeNbBits = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_nbNodesOnFirstLevelOfSubTreeNbBits;
	m_pDSI->HDESC_nbSubTreesChildrenNbBits = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_nbSubTreesChildrenNbBits;
	m_pDSI->HDESC_nbNodesOnLastLevelNbBits = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_nbNodesOnLastLevelNbBits;
	m_pDSI->HDESC_networkType = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_networkType;
	m_pDSI->HDESC_subTreeSizeNbBits = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_subTreeSizeNbBits;
	m_pDSI->HDESC_geometryNodesSizeNbBits = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->HDESC_geometryNodesSizeNbBits;

	// WSS
	m_pDSI->WSS_NbBitsIndex = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->WSS_NbBitsIndex;
	m_pDSI->WSS_NbBitsObjCount = ((SpacePartDecoderSpecificInfo*)pSpecificInfo)->WSS_NbBitsObjCount;

	return TRUE;
}

BOOL SpacePartDecoderLib :: DecodeSpacePartitionStream (LPBYTE pInput, int nInputLength)
{
	mbitstream* bs=new mbitstream(pInput);

	DecodeSpacePartitionMessage(bs);

	delete bs;
	return TRUE;
}

BOOL SpacePartDecoderLib :: DecodeFromBuffer (LPBYTE pInput, int nInputLength)
{
	mbitstream* bs=new mbitstream(pInput);

	// ANNE TODO: encode dsi in media, in case inband 
	DecodeSpecificInfo(bs);
	DecodeSpacePartitionMessage(bs);
	delete bs;
	return TRUE;
}


BOOL SpacePartDecoderLib :: DecodeSpecificInfo(mbitstream* bs)
{
	unsigned int type, NbBitsIndex, NbBitsCoef, NbBitsObjCount;

	bs->readnbit(type,8);
	m_pDSI->SPtype = type;
 
	switch (type) {
	case 0: // BSP
		m_pSPart->m_SPtype = 0;
		bs->readnbit(NbBitsIndex, 6);		// indexNbBits
		m_pDSI->BSP_NbBitsIndex = NbBitsIndex;
		bs->readnbit(NbBitsCoef, 6);		// coeff
		m_pDSI->BSP_NbBitsCoef = NbBitsCoef;
		bs->readnbit(NbBitsObjCount, 6);	// objcount
		m_pDSI->BSP_NbBitsObjCount = NbBitsObjCount;
		if (m_pSPart->m_BSPTree == NULL)
		{
			m_pSPart->m_BSPTree = new BSPTree();
		}
		break;

	case 1: // Cell-Portal
		m_pSPart->m_SPtype = 1;
		bs->readnbit(NbBitsIndex, 6);
		m_pDSI->CP_NbBitsCellIndex = NbBitsIndex;
		bs->readnbit(NbBitsIndex, 6);
		m_pDSI->CP_NbBitsPortalIndex = NbBitsIndex;
		bs->readnbit(NbBitsObjCount, 6);
		m_pDSI->CP_NbBitsCellCount = NbBitsObjCount;
		bs->readnbit(NbBitsObjCount, 6);
		m_pDSI->CP_NbBitsPortalCount = NbBitsObjCount;
		bs->readnbit(NbBitsObjCount, 6);
		m_pDSI->CP_NbBitsCellPortalCount = NbBitsObjCount;
		if (m_pSPart->m_CPPartition == NULL)
		{
			m_pSPart->m_CPPartition = new CPPartition();
		}
		break;

	case 2: // PVS
		m_pSPart->m_SPtype = 2;
		bs->readnbit(NbBitsIndex, 6);
		m_pDSI->PVS_NbBitsIndex = NbBitsIndex;
		bs->readnbit(NbBitsObjCount, 6);
		m_pDSI->PVS_NbBitsObjCount = NbBitsObjCount;
		break;
	
	case 3: // Footprint
		break;

	case 4: // WSS
		m_pSPart->m_SPtype = 4;
		if (m_pSPart->m_WSSPart == NULL)
		{
			m_pSPart->m_WSSPart = new WssPartition();
		}
		break;
	}

	return TRUE;
}


BOOL SpacePartDecoderLib :: DecodeSpacePartitionMessage(mbitstream* bs)
{
	bool bRet = false;

	switch (m_pDSI->SPtype)
	{
	case 0: // BSP
		m_pSPart->m_SPtype = 0;
		bRet = DecodeBSPMessage(bs);
		break;
	case 1: // cell portal
		m_pSPart->m_SPtype = 1;
		bRet = DecodeCellPortalMessage(bs);
		break;
	case 2: // PVS
		m_pSPart->m_SPtype = 2;
		bRet = DecodePVSMessage(bs);
		break;
	case 3: // footprint
		m_pSPart->m_SPtype = 3;
		bRet = DecodeHDescMessage(bs);
		break;
	case 4: //wss
		m_pSPart->m_SPtype = 4;
		bRet = DecodeWSSMessage(bs);
		break;
	}

	return bRet;
}


BOOL SpacePartDecoderLib :: DecodeBSPMessage(mbitstream* bs)
{
	unsigned int nBspCount;
	bs->readnbit(nBspCount, m_pDSI->BSP_NbBitsIndex);
	m_pSPart->m_BSPTree->m_nTotalNodes = nBspCount;

	for (unsigned int ibsp=0; ibsp<nBspCount; ibsp++)
	{
		unsigned int nindex, nparentindex, nfrontindex, noverlapindex, nbackindex;
		float a, b, c, d;

		bs->readnbit(nindex, m_pDSI->BSP_NbBitsIndex);
		bs->readnbit(nparentindex, m_pDSI->BSP_NbBitsIndex);	
		bs->readnbit(a, m_pDSI->BSP_NbBitsCoef);	// a
		bs->readnbit(b, m_pDSI->BSP_NbBitsCoef);	// b
		bs->readnbit(c, m_pDSI->BSP_NbBitsCoef);	// c
		bs->readnbit(d, m_pDSI->BSP_NbBitsCoef);	// d
		bs->readnbit(nfrontindex, m_pDSI->BSP_NbBitsIndex);
		bs->readnbit(noverlapindex, m_pDSI->BSP_NbBitsIndex);
		bs->readnbit(nbackindex, m_pDSI->BSP_NbBitsIndex);

		BSPNode* node = new BSPNode(nindex, nparentindex, a, b, c, d, nfrontindex, noverlapindex, nbackindex);

		// front list
		unsigned int nFrontObjCount;
		bs->readnbit(nFrontObjCount, m_pDSI->BSP_NbBitsObjCount);
		node->m_nFrontObjCount = nFrontObjCount;
		
		if (nFrontObjCount > 0)
			node->m_lFrontObjID = (sint32*)calloc(nFrontObjCount, sizeof(uint32));

		for (int iFront=0; iFront<nFrontObjCount; iFront++)
		{
			unsigned int nObjID;
			bs->readnbit(nObjID, m_pDSI->BSP_NbBitsIndex);
			node->m_lFrontObjID[iFront] = nObjID;
		}

		// overlap list
		unsigned int nOverlapObjCount;
		bs->readnbit(nOverlapObjCount, m_pDSI->BSP_NbBitsObjCount);
		node->m_nOverlapObjCount = nOverlapObjCount;

		if (nOverlapObjCount > 0)
			node->m_lOverlapObjID = (sint32*)calloc(nOverlapObjCount, sizeof(uint32));

		for (int iOverlap=0; iOverlap<nOverlapObjCount; iOverlap++)
		{
			unsigned int nObjID;
			bs->readnbit(nObjID, m_pDSI->BSP_NbBitsIndex);
			node->m_lOverlapObjID[iOverlap] = nObjID;
		}
		
		// back list
		unsigned int nBackObjCount;
		bs->readnbit(nBackObjCount, m_pDSI->BSP_NbBitsObjCount);
		node->m_nBackObjCount = nBackObjCount;

		if (nBackObjCount > 0)
			node->m_lBackObjID = (sint32*)calloc(nBackObjCount, sizeof(uint32));

		for (int iBack=0; iBack<nBackObjCount; iBack++)
		{
			unsigned int nObjID;
			bs->readnbit(nObjID, m_pDSI->BSP_NbBitsIndex);
			node->m_lBackObjID[iBack] = nObjID;
		}

		m_pSPart->m_BSPTree->AddNode(node);
	}

	return TRUE;	
}


BOOL SpacePartDecoderLib :: DecodeCellPortalMessage(mbitstream* bs)
{
	unsigned int msgtype;	// ANNE TODO: check flag is needed??? 
	unsigned int cellcount;

	// read header
	bs->readnbit(msgtype, 8);	// ANNE TODO: make adaptive
	bs->readnbit(cellcount, m_pDSI->CP_NbBitsCellCount);

	if (m_pSPart->m_CPPartition == NULL)
	{
		m_pSPart->m_CPPartition = new CPPartition();
	}

	m_pSPart->m_CPPartition->m_nCellCount = cellcount;

	// ANNE TODO: number of cells in stream, to be added in stream because it
	// could be different than total number of cells !!! 
	
	for (int icell=0; icell<cellcount; icell++)
	{
		unsigned int cellid;
		
		bs->readnbit(msgtype, 8);
		bs->readnbit(cellid, m_pDSI->CP_NbBitsCellIndex);

		// bounding box info
		float xcenter, ycenter, zcenter, dx, dy, dz;
		bs->readnbit(xcenter, 32);		// ANNE TODO: make adaptive
		bs->readnbit(ycenter, 32);		// ANNE TODO: make adaptive
		bs->readnbit(zcenter, 32);		// ANNE TODO: make adaptive
		bs->readnbit(dx, 32);			// ANNE TODO: make adaptive
		bs->readnbit(dy, 32);			// ANNE TODO: make adaptive
		bs->readnbit(dz, 32);			// ANNE TODO: make adaptive

		int nBytes = ceil((float)((float)cellcount/(float)8));
		char* cparray = new char[nBytes];
		memset(cparray, 0, nBytes);
		for (int ibyte=0; ibyte<nBytes; ibyte++)
		{
			unsigned short CpChar = 0x0000;
			bs->readnbit(CpChar, 8);
			cparray[ibyte] = CpChar;
		}

		CPCell* pCell = new CPCell();

		pCell->m_nID = cellid;
		pCell->m_xcenter = xcenter;
		pCell->m_ycenter = ycenter;
		pCell->m_zcenter = zcenter;
		pCell->m_dx = dx;
		pCell->m_dy = dy;
		pCell->m_dz = dz;
		pCell->m_bitarray = cparray;

		//m_pSPart->m_CPPartition->m_Cells.push_back(pCell);
		m_pSPart->m_CPPartition->AddCell(pCell);

	}	
	return true;
}


BOOL SpacePartDecoderLib :: DecodePVSMessage(mbitstream* bs)
{
	unsigned int msgtype;
	unsigned int objcount;
	unsigned int cellcount;
	unsigned int bRegular;
	unsigned int bCompressed;
	float zmin, zmax, xmin, xmax, ymin, ymax, dx, dy;

/*
	if (m_pSPart->m_PVSPart == NULL)
	{
		m_pSPart->m_PVSPart = new PvsPartition();
	}
*/
	// header // TODO check flag is needed??
	bs->readnbit(msgtype, 8);			// header flag
	bs->readnbit(objcount, 32/*m_pDSI->PVS_NbBitsObjCount*/);
	bs->readnbit(bRegular, 8); // ANNE TODO: one bit is enough!!
	bs->readnbit(bCompressed, 8); // ANNE TODO: one bit is enough!!
	bs->readnbit(zmin, 32); // ANNE TODO: should be adaptive
	bs->readnbit(zmax, 32); // ANNE TODO: should be adaptive

	

	if (bRegular)
	{
		// init reg pvs
		m_pSPart->m_PVSPart = new PvsPartitionRegular();
		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_nObjCount = objcount;
		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_IsRegular = true;
		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_IsCompressed = bCompressed;
		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_zmin = zmin;
		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_zmax = zmax;
		
		// read reg params
		bs->readnbit(xmin, 32); // // ANNE TODO: should be adaptive
		bs->readnbit(xmax, 32); // // ANNE TODO: should be adaptive
		bs->readnbit(ymin, 32); // // ANNE TODO: should be adaptive
		bs->readnbit(ymax, 32); // // ANNE TODO: should be adaptive
		bs->readnbit(dx, 32); // // ANNE TODO: should be adaptive
		bs->readnbit(dy, 32); // // ANNE TODO: should be adaptive

		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_xmin = xmin;
		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_xmax = xmax;
		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_ymin = ymin;
		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_ymax = ymax;
		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_dx = dx;
		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_dy = dy;
		int nx = (int)ceil((xmax-xmin)/dx);
		int ny = (int)ceil((ymax-ymin)/dy);
		((PvsPartitionRegular*)m_pSPart->m_PVSPart)->m_CellsByID = (PvsCell**)calloc(nx*ny, sizeof(PvsCell*));
	}

	bs->readnbit(cellcount, m_pDSI->PVS_NbBitsIndex);

	for (int icell=0; icell<cellcount; icell++)
	{
		// useless header
		unsigned h;
		bs->readnbit(h, 8); // ANNE TODO: check header is needed
		// read pvs array
		unsigned int ncellID;
		bs->readnbit(ncellID, m_pDSI->PVS_NbBitsIndex);
		int nBytes = ceil((float)((float)(m_pSPart->m_PVSPart)->m_nObjCount/8));
		char* pvs = new char[nBytes];
		memset(pvs, 0, nBytes);
		for (int ibyte=0; ibyte<nBytes; ibyte++)
		{
			unsigned short PvsChar = 0x0000;
			bs->readnbit(PvsChar, 8);
			pvs[ibyte] = PvsChar;
		}
		
		// add cell
		if ((m_pSPart->m_PVSPart)->m_IsRegular) {
			PvsCell* pCell = new PvsCellRegular();
			pCell->m_nID = ncellID;
			pCell->m_pvs = pvs;
			((PvsPartitionRegular*)m_pSPart->m_PVSPart)->AddPvsCell(pCell);
			((PvsPartitionRegular*)m_pSPart->m_PVSPart)->FillObjIDs(pCell);
		}
		else
		{
			PvsCell* pCell = new PvsCellIrregular();
			pCell->m_nID = ncellID;
			pCell->m_pvs = pvs;
			((PvsPartitionIrregular*)m_pSPart->m_PVSPart)->AddPvsCell(pCell);
		}
	}

//	bs->readnbit(cellcount, m_pDSI->PVS_NbBitsIndex);



	return TRUE;	
}


//BOOL SpacePartDecoderLib :: DecodeHDescMessage(mbitstream* bs)
//{
//	unsigned int wsstype;
/*		
	if (m_pSPart->m_HDescPart == NULL)
	{
		m_pSPart->m_HDescPart = new HDescPartition();
	}
	*/
/*
	// ANNE TODO: check flag is needed?? 
	bs->readnbit(msgtype, 8); 

	unsigned int nVolumes, nFaces;
	bs->readnbit(nVolumes, 32); // ANNE TODO: make adaptive
	bs->readnbit(nFaces, 32); // ANNE TODO: make adaptive

	for (int ivolume=0; ivolume<nVolumes; ivolume++)		
	{
		WssVolume* pVolume = new WssVolume();

		unsigned int nid;
		float centerX, centerY, centerZ, dX, dY, dZ;
		bs->readnbit(wsstype, 8); // ANNE TODO: make adaptive
		bs->readnbit(nid, 32); // ANNE TODO: make adaptive
		bs->readnbit(centerX, 32); // ANNE TODO: make adaptive
		bs->readnbit(centerY, 32); // ANNE TODO: make adaptive
		bs->readnbit(centerZ, 32); // ANNE TODO: make adaptive
		bs->readnbit(dX, 32); // ANNE TODO: make adaptive
		bs->readnbit(dY, 32); // ANNE TODO: make adaptive
		bs->readnbit(dZ, 32); // ANNE TODO: make adaptive

		pVolume->m_nID = nid;
		pVolume->m_centerX = centerX;
		pVolume->m_centerY = centerY;
		pVolume->m_centerZ = centerZ;
		pVolume->m_dX = dX;
		pVolume->m_dY = dY;
		pVolume->m_dZ = dZ;

		int wssbytes = ceil(nFaces/8);
		char* wssArray;
		wssArray = (char*)calloc(wssbytes, sizeof(char));
		for (int ibyte=0; ibyte<wssbytes; ibyte++)
		{
			unsigned short wsschar = 0x0000;
			bs->readnbit(wsschar, 8);
			wssArray[ibyte] = wsschar;
		}
		pVolume->m_wssSizeInBytes = wssbytes;
		pVolume->m_wssArray = wssArray;

		((WssPartition*)m_pSPart->m_WSSPart)->AddVolume(pVolume);
	}
*/
//	return true;
//}


BOOL SpacePartDecoderLib :: DecodeHDescMessage(mbitstream* bs)
//bool FPHDescDecodingChain::decodeHDescStream(uint8 *pInput, uint32 u4InputLength,
//											 HierarchicalDescriptionCM *pCM) {
{
	if (m_pSPart->m_HDescTree == NULL)
	{
		m_pSPart->m_HDescTree = new HDescTree();
	}

/*	if (NULL == pCM || NULL == pCM->getHLevelSelector()) return false;
	if (0 >= u4InputLength) return false;

	FPHLevelSelector *pLevelSelector= static_cast<FPHLevelSelector*>(pCM->getHLevelSelector());
	BitReader bs(pInput, u4InputLength);
*/
	// Read the number of sub-trees which are contained in the access unit.
	unsigned int u4NbSubTrees;
	bs->readnbit(u4NbSubTrees, m_pDSI->HDESC_nbSubTreesNbBits);

	unsigned int u4IdxSubTree, u4Index;
	unsigned int u4IdxOfFstNodeInSubTree, u4IdxOfParentOfFstN, u4NbNodesInSubTree, u4NbNodesOnFstLevel;
	unsigned int u4IdxNode, u4NbChildren, u4NbChildrenOfRootNotDecoded;
	/*sint32*/ unsigned int s4Radius;
	bool bIsDeltaXNeg, bIsDeltaYNeg, bIsDeltaZNeg, bIsDeltaRadiusNeg;
	unsigned int u1encodedMetricError, u1DeltaNbBits;
	float fMetricError, fGcX, fGcY, fGcZ, fParentGcX, fParentGcY, fParentGcZ, fSwap;


	HDescNode *pParentNode, *pNode;
	int u4NbChildrenToAdd;
	bool bParentLevelChanged;
//	ArrayOf<FPHDescNode*, 1> nodesToBeAdded;
//	std::vector<HDescNode*> nodesToBeAdded;
//	int u4NbChildrenToAdd;
	HDescNode *pBrother, *pFstNodeWithChildren;

	HDescNode** nodesToBeAdded = NULL;
	int arraysize=0;

	// Decode the sub-trees contained in the access unit.
	for (u4IdxSubTree= 0; u4IdxSubTree < u4NbSubTrees; u4IdxSubTree++) {
		// Read the header of the sub-tree.
		bs->readnbit(u4NbNodesInSubTree, m_pDSI->HDESC_nbNodesInSubTreeNbBits);
		bs->readnbit(u4NbNodesOnFstLevel, m_pDSI->HDESC_nbNodesOnFirstLevelOfSubTreeNbBits);
		bs->readnbit(u4IdxOfParentOfFstN, m_pDSI->HDESC_indexNbBits);
		bs->readnbit(u4Index, m_pDSI->HDESC_indexNbBits);
		u4IdxOfFstNodeInSubTree = u4Index;

		//nodesToBeAdded.resize(u4NbNodesInSubTree);
		// copy temp
		if (arraysize > 0)
		{
			HDescNode** temparray = (HDescNode**)calloc(arraysize, sizeof(HDescNode*));
			for (int i=0; i<arraysize; i++)
			{
				temparray[i] = nodesToBeAdded[i];
			}
			if (nodesToBeAdded != NULL) free(nodesToBeAdded); nodesToBeAdded = NULL;

			nodesToBeAdded = (HDescNode**)calloc(arraysize+u4NbNodesInSubTree, sizeof(HDescNode*));
			//memset(nodesToBeAdded, 0, arraysize+u4NbNodesInSubTree);
			for (int j=0; j<arraysize; j++)
			{
				nodesToBeAdded[j] = temparray[j];
			}
			free(temparray);
			arraysize += u4NbNodesInSubTree;
		}
		else
		{
			arraysize += u4NbNodesInSubTree;
			nodesToBeAdded = (HDescNode**)calloc(arraysize, sizeof(HDescNode*));
		}

		// Check if the parent node of the first node of the sub-tree is already in the hierarchical
		// description tree.

//		if (NULL == pLevelSelector->getRoot() && 0 == u4IdxOfFstNodeInSubTree) {
		if ( (m_pSPart->m_HDescTree->m_pRoot == NULL) && (0 == u4IdxOfFstNodeInSubTree) ) {
			// The sub-tree which is currently read is the root sub-tree. This sub-tree have to be the first
			// sub-tree sent (in all cases, it will always be the first sub-tree requested).
//			pParentNode= pLevelSelector->buildRoot();
			m_pSPart->m_HDescTree->AddRoot();
			pParentNode = m_pSPart->m_HDescTree->m_pRoot;
			pParentNode->SetNbMissingChildren(m_pDSI->HDESC_nbRootChildren);
			//u4NbChildrenOfRootNotDecoded= pParentNode->getNbMissingChildren();
			u4NbChildrenOfRootNotDecoded = m_pDSI->HDESC_nbRootChildren;
		}
		else {
			u4NbChildrenOfRootNotDecoded= 0;
			pParentNode= m_pSPart->m_HDescTree->GetNode(u4IdxOfParentOfFstN);
			if (NULL == pParentNode) {
//				_TRACE(L"The parent node of the node to add to the sub-tree has not been found\n");
				return false;
			}

			if (0 == pParentNode->GetNbMissingChildren()) {
//				_TRACE(L"Trying to add children to a node while this one is full.\n");
				return false;
			}

			pNode= m_pSPart->m_HDescTree->GetNode(u4IdxOfFstNodeInSubTree);
			if (NULL != pNode) {
//				_TRACE(L"Trying to add a sub-tree already present in the HDesc Tree\n");
				return false;
			}
		}
		
		// Read the nodes contained in the sub-tree and add them to the hierarchical description tree.
		pFstNodeWithChildren= NULL;
		for (u4IdxNode= 0; u4IdxNode < u4NbNodesInSubTree; u4IdxNode++) {
			bs->readnbit(u4NbChildren, m_pDSI->HDESC_nbChildrenNbBits);
			if (u4NbChildren > 0) 
			{
				// Decoding of the metric error.
				bs->readnbit(u1encodedMetricError, 8);
				
				if (0 == u1encodedMetricError)
				{
					fMetricError = m_pDSI->HDESC_minMetricError;
				}
				else 
				{
					double fExp = exp((double)((float)u1encodedMetricError * m_pDSI->HDESC_maxMetricErrorEncodingFunction / 255.0f));
					fMetricError = (float)pow(fExp - 1.0, 0.125);

					if (fMetricError < m_pDSI->HDESC_minMetricError)
					{
						fMetricError = m_pDSI->HDESC_minMetricError * 1.001f;
					}
				}
			}
			else fMetricError = 0.0f;

			if (u4NbChildrenOfRootNotDecoded > 0) {
				// We are decoding a node which will be one of the children of the root.
				bs->readnbit(fGcX, 32);	// float
				bs->readnbit(fGcY, 32); // float
				bs->readnbit(fGcZ, 32); // float

				// Applies a rotation of PI/2 along the x axis since in Origami, the z axis
				// is used as the depth dimension.
				fSwap= fGcY;
				fGcY= fGcZ;
				fGcZ= -fSwap;

//				s4Radius= static_cast<sint32>(bs->readnbit(m_pDSI->HDESC_rootChildrenRadiusNbBits));
				bs->readnbit(s4Radius, m_pDSI->HDESC_rootChildrenRadiusNbBits);

				u4NbChildrenOfRootNotDecoded--;
			}
			else
			{
				// The node is not a child of the root. Its GC coordinates are encoded thanks to a
				// subtraction between its own coordinates and the coordinates of its father.
				//u1DeltaNbBits= static_cast<uint8>(bs.readBits(5) + 1);
				bs->readnbit(u1DeltaNbBits, 5); u1DeltaNbBits++;
				bIsDeltaXNeg= bs->readbit();
				//fGcX= static_cast<float>(bs.readBits((uint8)u1DeltaNbBits)) * (bIsDeltaXNeg ? -1.0f : 1.0f);
				unsigned short ustemp;
				bs->readnbit(ustemp, u1DeltaNbBits); 
				fGcX = (float)ustemp;
				fGcX *= (bIsDeltaXNeg ? -1.0f : 1.0f);

				bIsDeltaYNeg= bs->readbit();
				//fGcY= static_cast<float>(bs.readBits((uint8)u1DeltaNbBits)) * (bIsDeltaYNeg ? -1.0f : 1.0f);
				bs->readnbit(ustemp, u1DeltaNbBits);
				fGcY = (float)ustemp; 
				fGcY *= (bIsDeltaYNeg ? -1.0f : 1.0f);
				
				bIsDeltaZNeg= bs->readbit();
				//fGcZ= static_cast<float>(bs.readBits((uint8)u1DeltaNbBits)) * (bIsDeltaZNeg ? -1.0f : 1.0f);
				bs->readnbit(ustemp, u1DeltaNbBits);
				fGcZ = (float)ustemp; 
				fGcZ *= (bIsDeltaZNeg ? -1.0f : 1.0f);
				
				// Applies a rotation of PI/2 along the x axis since in Origami, the z axis
				// is used as the depth dimension.
				fSwap= fGcY;
				fGcY= fGcZ;
				fGcZ= -fSwap;

				bIsDeltaRadiusNeg= bs->readbit();
				//s4Radius= static_cast<sint32>((bs.readBits((uint8)u1DeltaNbBits)) * (bIsDeltaRadiusNeg ? -1 : 1));
				bs->readnbit(ustemp, u1DeltaNbBits);
				s4Radius = (int)ustemp;
				s4Radius *= (bIsDeltaRadiusNeg ? -1.0f : 1.0f);
			}

			// Build a new node.
			HDescNode* pNode= new HDescNode(fGcX, fGcY, fGcZ, fMetricError, s4Radius);
			pNode->SetIndex(u4Index);
			pNode->SetNbMissingChildren(u4NbChildren);
			m_pSPart->m_HDescTree->AddNode(pNode);
			u4Index++;

			if (NULL == pFstNodeWithChildren && u4NbChildren > 0) pFstNodeWithChildren= pNode;

			nodesToBeAdded[u4IdxNode]= pNode;
		}


		// The nodes have been decoded. We add them to the hierarchical description tree.
		u4NbChildrenToAdd = pParentNode->GetNbMissingChildren();
		bParentLevelChanged= false;
		pBrother= NULL;
		for (u4IdxNode= 0; u4IdxNode < u4NbNodesInSubTree; u4IdxNode++) {
			while (u4NbChildrenToAdd < 1) {
				// We have to change the parent node.
				pParentNode= static_cast<HDescNode*>(pParentNode->GetBrother());
				if (!bParentLevelChanged && 0 == u4NbNodesOnFstLevel) {
					bParentLevelChanged= true;
					pParentNode= pFstNodeWithChildren;
				}

				if (NULL == pParentNode) {
/*					for (; u4IdxNode < u4NbNodesInSubTree; u4IdxNode++) delete nodesToBeAdded[u4IdxNode];
					_TRACE(L"No parent found to add the remaining children of the sub-tree.\n");
*/					return false;
				}

				u4NbChildrenToAdd= pParentNode->GetNbMissingChildren();
			}

			pNode= nodesToBeAdded[u4IdxNode];
			if (pNode->GetIndex() < pParentNode->GetIndex()) {
				// Property not respected: a child has always an index greater than its father index.
/*				_TRACE(L"idx node %d < idx node %d while node %d is parent of %d",
					   pNode->getIndex(), pParentNode->getIndex(), pParentNode->getIndex(), pNode->getIndex());
				for (; u4IdxNode < u4NbNodesInSubTree; u4IdxNode++) delete nodesToBeAdded[u4IdxNode];
*/				return false;
			}
			
			if (!pParentNode->IsRoot()) {
				pParentNode->GetGCCoordinates(fParentGcX, fParentGcY, fParentGcZ);
				pNode->GetGCCoordinates(fGcX, fGcY, fGcZ);
				fGcX= fGcX * m_pDSI->HDESC_acquisitionPrecision + fParentGcX;
				fGcY= fGcY * m_pDSI->HDESC_acquisitionPrecision + fParentGcY;
				fGcZ= fGcZ * m_pDSI->HDESC_acquisitionPrecision + fParentGcZ;
				pNode->SetGCCoordinates(fGcX, fGcY, fGcZ);

				s4Radius= pParentNode->GetRadius() - pNode->GetRadius();

				if (s4Radius < 0) {
/*					_TRACE(L"Negative radius for node %d parent %d (radius: %d)\n",
						   pNode->GetIndex(), pParentNode->GetIndex(), s4Radius);
*/					s4Radius= 0;
				}

				pNode->SetRadius(s4Radius);
			}

			if (!bParentLevelChanged) u4NbNodesOnFstLevel--;
			pParentNode->AddChild(pNode);
			pNode->SetParent(pParentNode);
			pNode->SetLevel(pParentNode->GetLevel() + 1);

			if (NULL != pBrother) pBrother->SetBrother(pNode);
			pBrother= pNode;

			u4NbChildrenToAdd--;
		}
	}

	if (nodesToBeAdded != NULL) free(nodesToBeAdded); nodesToBeAdded = NULL;

/*
	int nnodes = m_pSPart->m_HDescTree->m_SortedNodes.size();
	SORTED_NODES::iterator itnodes;
	for (itnodes=m_pSPart->m_HDescTree->m_SortedNodes.begin(); itnodes!=m_pSPart->m_HDescTree->m_SortedNodes.end(); itnodes++)
	{
		std::pair<class HDescNode * const,int> pPair = *itnodes;
		HDescNode* pNode = pPair.first;
		int i = pPair.second;
	}
*/
	return true;
}



BOOL SpacePartDecoderLib :: DecodeWSSMessage(mbitstream* bs)
{
	unsigned int wsstype;
	
	
	if (m_pSPart->m_WSSPart == NULL)
	{
		m_pSPart->m_WSSPart = new WssPartition();
	}

	// ANNE TODO: check flag is needed?? 
	bs->readnbit(wsstype, 8); 

	unsigned int nVolumes, nFaces;
	bs->readnbit(nVolumes, 32); // ANNE TODO: make adaptive
	bs->readnbit(nFaces, 32); // ANNE TODO: make adaptive

	for (int ivolume=0; ivolume<nVolumes; ivolume++)		
	{
		WssVolume* pVolume = new WssVolume();

		unsigned int nid;
		float centerX, centerY, centerZ, dX, dY, dZ;
		bs->readnbit(wsstype, 8); // ANNE TODO: make adaptive
		bs->readnbit(nid, 32); // ANNE TODO: make adaptive
		bs->readnbit(centerX, 32); // ANNE TODO: make adaptive
		bs->readnbit(centerY, 32); // ANNE TODO: make adaptive
		bs->readnbit(centerZ, 32); // ANNE TODO: make adaptive
		bs->readnbit(dX, 32); // ANNE TODO: make adaptive
		bs->readnbit(dY, 32); // ANNE TODO: make adaptive
		bs->readnbit(dZ, 32); // ANNE TODO: make adaptive

		pVolume->m_nID = nid;
		pVolume->m_centerX = centerX;
		pVolume->m_centerY = centerY;
		pVolume->m_centerZ = centerZ;
		pVolume->m_dX = dX;
		pVolume->m_dY = dY;
		pVolume->m_dZ = dZ;

		int wssbytes = ceil(nFaces/8);
		char* wssArray;
		wssArray = (char*)calloc(wssbytes, sizeof(char));
		for (int ibyte=0; ibyte<wssbytes; ibyte++)
		{
			unsigned short wsschar = 0x0000;
			bs->readnbit(wsschar, 8);
			wssArray[ibyte] = wsschar;
		}
		pVolume->m_wssSizeInBytes = wssbytes;
		pVolume->m_wssArray = wssArray;

		((WssPartition*)m_pSPart->m_WSSPart)->AddVolume(pVolume);
	}

	return true;
}